@@ -226,7 +226,7 @@ rdfe_load(File, Options) :-
     rdfe_current_transaction(TID),
     absolute_file_name(File,
                        [ access(read),
-                         extensions([rdf,rdfs,owl,''])
+                         extensions([rdf,rdfs,owl,ttl,nt,''])
                        ], Path),
     rdf_load(Path,
              [ graph(Graph),
@@ -1072,7 +1072,7 @@ replay_actions([H|T]) :-
 %   appears changed, but this isn't really   easy  as copying and OS
 %   differences makes it hard to decide on changes by length as well
 %   as modification time. Alternatively we could   save the state in
-%   seperate quick-load states.
+%   separate quick-load states.
 
 replay_action(retract(_, Subject, Predicate, Object, PayLoad)) :-
     rdf_retractall(Subject, Predicate, Object, PayLoad).
